package auth

import (
	"context"
	"fmt"
	"path"
	"sync"

	"git.loafle.net/commons/configuration-go"
	cdr "git.loafle.net/commons/di-go/registry"
	logging "git.loafle.net/commons/logging-go"
	occa "git.loafle.net/overflow/commons-go/core/annotation"
	occi "git.loafle.net/overflow/commons-go/core/interfaces"
	ocnc "git.loafle.net/overflow/commons-go/noauthprobe/config"
	"git.loafle.net/overflow/probe/config"

	"git.loafle.net/overflow/probe/auth/service"
)

type Authenticator struct {
	Config    *config.Config
	ConfigDir string

	authConfig ocnc.Auth

	services []interface{}

	stopChan chan struct{}
	stopWg   sync.WaitGroup
}

func (a *Authenticator) EndableStart() (<-chan error, error) {
	if a.stopChan != nil {
		return nil, fmt.Errorf("authenticator already running. Stop it before starting it again")
	}

	authConfigPath := path.Join(a.ConfigDir, ocnc.ConfigFileName)
	conf := configuration.New()

	if configuration.Exists(authConfigPath) {
		if err := conf.Load(&a.authConfig, authConfigPath); nil != err {
			logging.Logger().Errorf("%s %v", err)
			return nil, fmt.Errorf("loading of auth config file[%s] failed", authConfigPath)
		}
	}

	if nil != a.authConfig.DeniedDate {
		return nil, fmt.Errorf("cannot start because this probe have been denied from overFlow at %s", a.authConfig.DeniedDate.String())
	}

	cdr.RegisterResource("AuthConfig", &a.authConfig)
	authDoneChan := make(chan error)
	cdr.RegisterResource("AuthDoneChan", authDoneChan)

	services, err := cdr.GetInstancesByAnnotationType(occa.RPCServiceAnnotationType)
	if nil != err {
		return nil, err
	}
	a.services = services

	if err := occi.ExecServices(a.services, occi.ServiceMethodInit, service.OrderedServices, false); nil != err {
		return nil, err
	}

	endChan := make(chan error)
	a.stopChan = make(chan struct{})

	a.stopWg.Add(1)
	go a.handleAuthenticator(authDoneChan, endChan)

	return endChan, nil
}

func (a *Authenticator) Stop(ctx context.Context) error {
	if a.stopChan == nil {
		return fmt.Errorf("Authenticator: must be started before stopping it")
	}
	close(a.stopChan)
	a.stopWg.Wait()

	if err := occi.ExecServices(a.services, occi.ServiceMethodDestroy, service.OrderedServices, true); nil != err {
		return err
	}

	a.stopChan = nil

	return nil
}

func (a *Authenticator) logHeader() string {
	return "Authenticator:"
}

func (a *Authenticator) handleAuthenticator(authDoneChan chan error, endChan chan<- error) {
	var err error
	defer func() {
		a.stopWg.Done()
		endChan <- err
	}()

	err = occi.ExecServices(a.services, occi.ServiceMethodStart, service.OrderedServices, false)
	if nil != err {
		return
	}

LOOP:
	for {
		select {
		case err = <-authDoneChan:
			break LOOP
		case <-a.stopChan:
			break LOOP
		}
	}

	err = occi.ExecServices(a.services, occi.ServiceMethodStop, service.OrderedServices, true)
}
